from sqlalchemy import create_engine, exc

from nomad.engine import BaseEngine, DBError


class SAEngine(BaseEngine):
    def connect(self):
        return create_engine(self.url)

    def prepare(self, statement):
        if self.connection.name in ('mysql', 'pgsql', 'postgresql'):
            return statement.replace('?', '%s')
        return statement

    def query(self, statement, *args, **kwargs):
        statement = self.prepare(statement)
        try:
            return self.connection.execute(statement, *args, **kwargs)
        except exc.SQLAlchemyError, e:
            raise DBError(str(e))

    begin = rollback = commit = lambda self: None


engine = SAEngine



import sys
import re
import os, os.path as op
import shlex
import imp
import json
from subprocess import Popen, PIPE
from configparser import ConfigParser, ExtendedInterpolation

from termcolor import cprint

if sys.version_info[0] == 3:
    shell_split = shlex.split
else:
    def shell_split(cmd):
        return shlex.split(cmd.encode('utf-8'))

NUM_RE = re.compile('(\d+)')


class NomadError(Exception):
    pass

class NomadIniNotFound(Exception):
    pass


def cachedproperty(f):
    """Returns a cached property that is calculated by function f
    """
    def get(self):
        try:
            return self._property_cache[f]
        except AttributeError:
            self._property_cache = {}
        except KeyError:
            pass
        x = self._property_cache[f] = f(self)
        return x

    return property(get)


def abort(msg, code=1):
    cprint('Error: %s' % msg, 'red', file=sys.stderr)
    sys.exit(code)


def shsplit(s):
    if not isinstance(s, str):
        s = s.encode('utf-8')
    return shlex.split(s)


def humankey(fn):
  '''Turn a string into a list of substrings and numbers.

  This can be used as a key function for ``sorted``::

    >>> s = lambda *x: list(sorted(x, key=humankey))
    >>> print s('up-1', 'up-5', 'up-15', 'up-50')
    ['up-1', 'up-5', 'up-15', 'up-50']
    >>> print s('up-1.sql', 'up.sql', 'up1.sql')
    ['up.sql', 'up1.sql', 'up-1.sql']
    >>> print s('up.rb', 'up.py') # check extension sorting
    ['up.py', 'up.rb']
  '''
  fn, ext = os.path.splitext(fn)
  return [int(s) if s.isdigit() else s for s in NUM_RE.split(fn)], ext


def loadpath(path):
    modname = 'nomad_url_%s' % path.replace('/', '_').replace('.', '_')
    path = op.expanduser(op.expandvars(path))
    if op.isdir(path):
        # path/__init__.py
        d, f = op.split(path.rstrip('/'))
        fd, fpath, desc = imp.find_module(f, [d])
        return imp.load_module(modname, fd, fpath, desc)
    else:
        return imp.load_source(modname, path)


def clean_sql(sql):
    return '\n'.join(x for x in sql.split('\n')
                     if not x.strip().startswith('--'))

### URL retrievers

def get_python(path):
    pypath, attr = path.split(':')
    if '/' in pypath or pypath.endswith('.py'):
        # load from file
        mod = loadpath(pypath)
    else:
        # load from sys.path
        mod = __import__(pypath, {}, {}, [''])
    try:
        return reduce(lambda x, y: getattr(x, y), attr.split('.'), mod)
    except AttributeError:
        raise AttributeError("No attr '%s' in module %s" % (attr, pypath))


def get_file(path):
    return open(path).read().strip()


def get_command(cmd):
    p1 = Popen(shell_split(cmd), stdout=PIPE)
    return p1.communicate()[0].strip().decode('utf-8')


def get_json(path):
    fn, path = path.split(':')
    obj = json.load(open(fn))
    path = map(lambda x: int(x) if x.isdigit() else x, path.split('.'))
    return reduce(lambda x, y: x[y], path, obj)


def get_ini(path):
    fn, path = path.split(':')
    section, key = path.split('.')
    cfg = ConfigParser(interpolation=ExtendedInterpolation())
    cfg.read([fn])
    try:
        return cfg[section][key]
    except KeyError:
        raise KeyError('%s not found in %s' % (path, fn))


def get_yaml(path):
    try:
        import yaml
    except ImportError:
        abort('Please, install PyYAML to parse YAML config.')
    fn, path = path.split(':')
    obj = yaml.load(open(fn))
    path = map(lambda x: int(x) if x.isdigit() else x, path.split('.'))
    return reduce(lambda x, y: x[y], path, obj)


def get_env(key):
    return os.environ[key]


URLTYPES = {
    'python': get_python,
    'py': get_python,
    'file': get_file,
    'command': get_command,
    'cmd': get_command,
    'json': get_json,
    'ini': get_ini,
    'yaml': get_yaml,
    'env': get_env,
    }


def geturl(urlspec):
    for url in shsplit(urlspec):
        bits = url.split(':', 1)
        if len(bits) > 1 and bits[0] in URLTYPES:
            try:
                url = URLTYPES[bits[0]](bits[1])
                if url:
                    return url
            except (IOError, OSError, KeyError):
                pass
        else:
            return url
    abort("could not read database url from spec '%s'" % urlspec)


if __name__ == '__main__':
    import doctest
    doctest.testmod()



import os, os.path as op
from datetime import datetime
from configparser import ConfigParser, ExtendedInterpolation
from subprocess import call
from functools import wraps

from nomad.utils import (cachedproperty, geturl, NomadError, NomadIniNotFound,
                         clean_sql, abort, humankey)
from nomad.engine import DBError


def tx(getrepo):
    def decorator(f):
        @wraps(f)
        def inner(self, *args, **kwargs):
            repo = getrepo(self)
            try:
                repo.engine.begin()
                f(self, *args, **kwargs)
                repo.engine.commit()
            except:
                repo.engine.rollback()
                raise
        return inner
    return decorator


class Repository(object):
    DEFAULTS = {
        'nomad': {'table': 'nomad'},
        }

    def __init__(self, confpath, overrides=None):
        self.conf = ConfigParser(
            interpolation=ExtendedInterpolation(),
            defaults={
                'confpath': op.abspath(confpath),
                'confdir': op.dirname(op.abspath(confpath)),
            })
        self.conf.read_dict(self.DEFAULTS)
        if not self.conf.read([confpath]):
            raise NomadIniNotFound(confpath)

        for k, v in (overrides or {}).iteritems():
            section, key = k.split('.')
            self.conf.set(section, key, v)

        self.confpath = confpath
        self.path = self.conf.get('nomad', 'path',
                                  fallback=op.dirname(confpath) or '.')

        try:
            enginepath = self.conf['nomad']['engine']
            if not '.' in enginepath:
                enginepath = 'nomad.engine.' + enginepath
        except KeyError:
            raise NomadError('nomad.engine is not defined in config file')

        try:
            enginemod = __import__(enginepath, {}, {}, [''])
        except ImportError, e:
            raise NomadError('cannot use engine %s: %s' % (enginepath, e))

        try:
            self.url = geturl(self.conf['nomad']['url'])
        except KeyError:
            abort('database url in %s is not found' % self)

        self.engine = getattr(enginemod, 'engine')(self.url)
        try:
            self.engine.connection
        except DBError, e:
            abort(e)

    def __repr__(self):
        return '<%s: %s>' % (type(self).__name__, self.path)

    def get(self, name):
        if name.endswith('/'):
            name = name[:-1]
        applied = name in self.appliednames
        return Migration(self, name, applied=applied)

    # actual work is done here
    @tx(lambda self: self)
    def init_db(self):
        self.engine.init(self.conf['nomad']['table'])

    @cachedproperty
    def available(self):
        migrations = [self.get(x) for x in os.listdir(self.path) if
                      op.isdir(op.join(self.path, x))]
        return list(sorted(migrations))

    @cachedproperty
    def appliednames(self):
        return [x for (x, ) in
                self.engine.query('SELECT name FROM %s ORDER BY date' %
                                  self.conf['nomad']['table'])]

    @property
    def applied(self):
        return [self.get(x) for x in self.appliednames]

    def get_env(self):
        return dict(('NOMAD_' + k.upper(), v)
                    for k, v in self.conf['nomad'].items())


class Migration(object):
    SINGLETONS = {}

    def __new__(cls, repo, name, **kwargs):
        key = (repo, name)
        if key not in cls.SINGLETONS:
            cls.SINGLETONS[key] = object.__new__(cls)
        return cls.SINGLETONS[key]

    def __init__(self, repo, name, applied=False):
        self.repo = repo
        self.name = name
        self.conf = ConfigParser(
            interpolation=ExtendedInterpolation(),
            defaults={
                'confpath': op.abspath(self.repo.confpath),
                'confdir': op.dirname(op.abspath(self.repo.confpath)),
                'dir': op.abspath(op.join(repo.path, name))
            })
        self.conf.read([op.join(repo.path, name, 'migration.ini')])
        deps = self.conf.get('nomad', 'dependencies', fallback='').split(',')
        self._deps = [x.strip() for x in deps if x.strip()]

        self.applied = applied
        self.exists = op.exists(op.join(repo.path, name))

    def __repr__(self):
        return '<%s: %s>' % (type(self).__name__, str(self))

    def __str__(self):
        return self.name

    def __lt__(self, other):
        if isinstance(other, Migration) and self.repo == other.repo:
            return humankey(self.name) < humankey(other.name)
        raise TypeError('Migrations can be compared only with other migrations')

    @property
    def path(self):
        return op.join(self.repo.path, self.name)

    @cachedproperty
    def dependencies(self):
        return map(self.repo.get, self._deps)

    def _apply(self, env=None):
        '''The real work for applying a migration
        '''
        filenames = sorted(os.listdir(self.path), key=humankey)

        for fn in filenames:
            if fn == 'migration.ini':
                continue
            path = op.join(self.path, fn)

            if fn.endswith('.sql'):
                with open(path) as f:
                    self.repo.engine.query(clean_sql(f.read()))
                print '  sql migration applied: %s' % fn

            elif os.access(path, os.X_OK):
                callenv = dict(os.environ,
                               # for backward compatibility
                               NOMAD_DBURL=self.repo.url,
                               **self.repo.get_env())
                if env:
                    callenv.update(env)
                retcode = call(path, env=callenv)
                if retcode:
                    raise DBError('script failed: %s' % fn)
                print '  script migration applied: %s' % fn

            else:
                print '  skipping file: %s' % fn

    @tx(lambda self: self.repo)
    def apply(self, env=None, fake=False):
        '''Apply a migration
        '''
        for dep in self.dependencies:
            if not dep.applied:
                dep.apply(env=env, fake=fake)

        if not fake:
            print 'applying migration %s:' % self
            self._apply(env=env)
        else:
            print 'applying "fake" migration %s' % self

        self.repo.engine.query('INSERT INTO %s (name, date) VALUES (?, ?)'
                               % self.repo.conf['nomad']['table'],
                               self.name, str(datetime.now()))
        self.applied = True



class BaseEngine(object):
    _connection = None

    def __init__(self, url):
        self.url = url

    def __repr__(self):
        return '<%s: %s>' % (type(self).__name__, self.url)

    @property
    def connection(self):
        if not self._connection:
            self._connection = self.connect()
        return self._connection

    def connect(self):
        raise NotImplementedError()

    def query(self, statement, *args, **kwargs):
        raise NotImplementedError()

    def begin(self):
        self.connection.begin()

    def commit(self):
        self.connection.commit()

    def rollback(self):
        self.connection.rollback()

    @property
    def datetime_type(self):
        if self.url.startswith('pgsql') or self.url.startswith('postgresql'):
            return 'timestamp'
        return 'datetime'

    def init(self, tablename):
        self.query('''CREATE TABLE %s (
            name varchar(255) NOT NULL,
            date %s NOT NULL
        )''' % (tablename, self.datetime_type))


class DBError(Exception):
    '''Database exception in nomad'''



#!/usr/bin/env python

import sys, os, re
from setuptools import setup, find_packages


DEPS = ['opster>=4.0', 'termcolor']
extra = {}

if sys.version_info[0] >= 3:
    extra.update(dict(
        use_2to3=True,
        convert_2to3_doctests=['nomad/utils.py'],
    ))
else:
    DEPS.append('configparser')


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

def find_version():
    val, = re.findall(r"__version__ = '([^']+)'",
                      read('nomad/__init__.py'))
    return val


config = dict(
    name = 'nomad',
    description = 'simple sql migration tool to save you from becoming mad',
    long_description = read('README.rst'),
    license = 'ISC',
    version = find_version(),
    author = 'Alexander Solovyov',
    author_email = 'alexander@solovyov.net',
    url = 'http://github.com/piranha/nomad/',
    classifiers = [
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development',
        'Topic :: Software Development :: Version Control',
        'Topic :: Database'
        ],

    install_requires = DEPS,
    packages = find_packages(),
    entry_points = {'console_scripts': ['nomad=nomad:app.dispatch']},
    platforms='any',
    **extra)

if __name__ == '__main__':
    setup(**config)



import urlparse
import urllib

from nomad.engine import BaseEngine, DBError


def path2dict(p, **renames):
    '''Convert urlparse result to dict
    '''
    result = {}
    for k in 'hostname port path username password'.split():
        value = getattr(p, k, None)
        if not value:
            continue
        if isinstance(value, str):
            value = urllib.unquote(value)
        if k == 'path':
            value = value.lstrip('/')
        result[renames.get(k, k)] = value
    return result


class Connection(object):
    exc = None

    def __init__(self):
        self.connection = self.connect()

    def connect(self):
        raise NotImplementedError()

    def prepare(self, statement):
        return statement

    def fetch(self, cursor):
        return cursor.fetchall()

    def query(self, statement, *args):
        statement = self.prepare(statement)
        c = self.connection.cursor()
        try:
            c.execute(statement, args)
        except self.exc, e:
            raise DBError(e)

        data = self.fetch(c)
        c.close()
        return data

    def begin(self):
        pass

    def commit(self):
        return self.connection.commit()

    def rollback(self):
        return self.connection.rollback()


class Sqlite(Connection):
    _conn = None

    def __init__(self, path):
        # [1:] strips first leading slash here
        self.path = path.path[1:]
        import sqlite3
        self.module = sqlite3
        self.exc = sqlite3.Error
        Connection.__init__(self)

    def connect(self):
        return self.module.connect(self.path)


class Mysql(Connection):
    _conn = None
    def __init__(self, path):
        self.parameters = path2dict(path, hostname='host',
                                    password='passwd', path='db', username='user')
        import MySQLdb
        self.module = MySQLdb
        self.exc = MySQLdb.MySQLError
        Connection.__init__(self)

    def connect(self):
        return self.module.connect(**self.parameters)

    def prepare(self, statement):
        return statement.replace('?', '%s')

    def begin(self):
        return self.query('BEGIN')

    def commit(self):
        return self.query('COMMIT')

    def rollback(self):
        return self.query('ROLLBACK')


class Pgsql(Connection):
    _conn = None
    def __init__(self, path):
        self.parameters = path2dict(path, hostname='host', path='database', username='user')
        import psycopg2
        self.module = psycopg2
        self.exc = psycopg2.Error
        Connection.__init__(self)

    def connect(self):
        try:
            return self.module.connect(**self.parameters)
        except self.module.OperationalError, e:
            raise DBError('psycopg2: %s' % e)

    def prepare(self, statement):
        return statement.replace('?', '%s')

    def fetch(self, cursor):
        if cursor.rowcount == -1:
            return []
        try:
            return cursor.fetchall()
        except self.module.ProgrammingError:
            return []


CONNECTORS = {'sqlite': Sqlite, 'mysql': Mysql,
              'pgsql': Pgsql, 'postgresql': Pgsql, 'postgres': Pgsql}


class DBEngine(BaseEngine):
    def connect(self):
        p = urlparse.urlparse(self.url)
        if p.scheme not in CONNECTORS:
            raise DBError('scheme "%s" not supported' % p.scheme)
        return CONNECTORS[p.scheme](p)

    def query(self, statement, *args):
        return self.connection.query(statement, *args)


engine = DBEngine



#!/usr/bin/env python

import os, os.path as op, sys
from datetime import date

from opster import Dispatcher
from termcolor import cprint, colored

from nomad.repo import Repository
from nomad.engine import DBError
from nomad.utils import abort, NomadError, NomadIniNotFound


__version__ = '1.12'


GLOBAL = [
    ('c', 'config', 'nomad.ini', 'path to config file'),
    ('D', 'define', {}, 'override config values'),
    ]


EXAMPLE_INI = '''
  [nomad]
  engine = dbapi
  url = sqlite://data.db
'''


def getconfig(func):
    if func.__name__.startswith('help') or func.__name__ in ('version',):
        return func
    def inner(*args, **kwargs):
        try:
            repo = Repository(kwargs['config'], kwargs['define'])
        except NomadIniNotFound, e:
            sys.stderr.write("Create '%s' to use nomad, example:\n%s\n" %
                             (e, EXAMPLE_INI))
            abort('config file not found')
        except (IOError, NomadError), e:
            abort(e)

        return func(repo=repo, *args, **kwargs)
    return inner

app = Dispatcher(globaloptions=GLOBAL, middleware=getconfig)


@app.command()
def init(**opts):
    '''Initialize database migration management
    '''
    try:
        opts['repo'].init_db()
    except DBError, e:
        abort(e)
    print 'Versioning table initialized successfully'


@app.command(name='list', aliases=('ls',))
def list_(all=('a', False, 'show all migrations (default: only non-applied)'),
         **opts):
    '''List migrations
    '''
    repo = opts['repo']
    all_migrations = repo.available + [m for m in repo.applied
                                       if m not in repo.available]
    for m in all_migrations:
        if m not in repo.available:
            print colored(m, 'red') + ' (not on disk)'
        elif m in repo.applied:
            if all:
                cprint(m, 'magenta')
        else:
            out = colored(m, 'green')
            deps = []
            for dep in m.dependencies:
                if dep not in repo.applied:
                    deps.append(dep)
            if deps:
                out += ' (%s)' % ', '.join(map(str, deps))
            print out


@app.command()
def create(name,
           dependencies=('d', [], 'migration dependencies'),
           prefix_date=('p', False, 'prefix migration name with date'),
           **opts):
    '''Create new migration
    '''
    repo = opts['repo']
    deps = map(repo.get, dependencies)

    if prefix_date:
        name = date.today().strftime('%Y%m%d-') + name

    path = op.join(repo.path, name)
    try:
        os.mkdir(path)
    except OSError, e:
        if e.errno == 17:
            abort('directory %s already exists' % path)
        raise

    with open(op.join(path, 'migration.ini'), 'w') as f:
        f.write('[nomad]\n')
        f.write('dependencies = %s\n' % ', '.join(d.name for d in deps))
    with open(op.join(path, 'up.sql'), 'w') as f:
        f.write('-- SQL ALTER statements for database migration\n')


@app.command()
def apply(all=('a', False, 'apply all available migrations'),
          init=('', False, 'init if not initialized yet'),
          env=('e', [], 'list of additional environment variables'),
          fake=('', False, 'record migration as applied, but do not do anything'),
          *names,
          **opts):
    '''Apply migration and all of it dependencies

    You can pass additional environment variables to your migrations scripts
    using option `-e` (in addition to `NOMAD_<confvar>`s, which are passed by
    default):

        nomad apply -e ONE=one -e TWO=two -a

    '''
    repo = opts['repo']
    if init:
        try:
            repo.init_db()
        except DBError:
            pass

    if names:
        migrations = map(repo.get, names)
    elif all:
        migrations = [x for x in repo.available if x not in repo.applied]
    else:
        abort('Supply names of migrations to apply')

    if env:
        env = dict(x.split('=', 1) for x in env)

    for m in migrations:
        if m.applied:
            abort('migration %s is already applied' % m)

    for m in migrations:
        try:
            m.apply(env=env, fake=fake)
        except DBError, e:
            abort('cannot apply migration %s: %s' % (m, e))


@app.command()
def info(**opts):
    '''Show information about repository
    '''
    repo = opts['repo']
    print '%s:' % repo
    print '  %s' % repo.engine
    try:
        print '  %s applied' % len(repo.applied)
        print '  %s unapplied' % (len(repo.available) - len(repo.applied))
    except DBError:
        print '  Uninitialized repository'


@app.command()
def version(**opts):
    '''Show nomad version
    '''
    print 'Nomad v%s' % __version__


if __name__ == '__main__':
    app.dispatch()
